from itertools import chain
from typing import Any, List, Tuple, Union


class Tumbler:
    def __init__(self, dice: List) -> None:
        self.callables = list(filter(lambda d: callable(d), dice))
        self.ints = list(filter(lambda d: isinstance(d, int), dice))

    def __call__(self) -> Tuple[Any, ...]:
        raise NotImplementedError()

    def __rmul__(self, other_value: Union[Any]) -> Any:
        return MulTumbler([self for i in range(other_value)])

    def __mul__(self, other_value: Union[Any]) -> Any:
        return MulTumbler([self for i in range(other_value)])

    def __add__(self, other_value: Union[Any]) -> Any:
        return SumTumbler([self, other_value])


class SumTumbler(Tumbler):
    def __init__(self, dice: List) -> None:
        super().__init__(dice)

        if len(self.ints):
            raise TypeError("Cannot use integers in SumTumbler")

    def __call__(self) -> Tuple[Any, ...]:
        def is_a_tumbler(d: Any) -> bool:
            return isinstance(d, SumTumbler) or isinstance(d, MulTumbler)

        dice = [c for c in self.callables if not is_a_tumbler(c)]
        tumblers = [c for c in self.callables if is_a_tumbler(c)]

        dice_results = tuple([d() for d in dice])
        tumbler_results = tuple(chain.from_iterable([t() for t in tumblers]))
        return dice_results + tumbler_results


class MulTumbler(Tumbler):
    def __init__(self, dice: List) -> None:
        super().__init__(dice)

    def __call__(self) -> Tuple[Any, ...]:
        if self.ints and self.callables:
            results = tuple([self.callables[0]() for _ in range(self.ints[0])])
        else:
            results = tuple(chain.from_iterable([c() for c in self.callables]))
        return results



import random
from functools import partial
from typing import Any, Callable, Union

from dragn.dice.tumbler import MulTumbler, SumTumbler, Tumbler


class Die:
    def __init__(self, max_value: int, randomness: Callable) -> None:
        self.randomness = randomness
        self.max_value = max_value

    def __call__(self) -> Any:
        return self.randomness(1, self.max_value)


class DieBuilder:
    def __init__(self, max_value: int) -> None:
        self.the_die = partial(self._die_builder, max_value)
        self.max_value = max_value

    @staticmethod
    def _die_builder(max_value: int, randomness: Callable = random.randint) -> Any:
        return Die(max_value, randomness)()

    def __call__(self) -> Any:
        return self.the_die()

    def __mul__(self, value: Union["DieBuilder", int, Tumbler]) -> Tumbler:
        return self._multiply(value)

    def __rmul__(self, value: Union["DieBuilder", int, Tumbler]) -> Tumbler:
        return self._multiply(value)

    def _multiply(self, other_value: Union["DieBuilder", int, Tumbler]) -> Tumbler:
        return MulTumbler([self, other_value])

    def __add__(self, value: Union["DieBuilder", int, Tumbler]) -> Tumbler:
        return self._add(value)

    def __radd__(self, value: Union["DieBuilder", int, Tumbler]) -> Tumbler:
        return self._add(value)

    def _add(self, other_value: Union["DieBuilder", int, Tumbler]) -> Tumbler:
        return SumTumbler([self, other_value])

    def __str__(self) -> str:
        return f"D{self.max_value}"



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    # Project
    name="dragn",
    version="0.3.1",
    description="A library to emulate rolling dice",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lurst/dragn",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
    ],
    # Author
    author="Gil Goncalves",
    author_email="lursty@gmail.com",
    # Code
    packages=setuptools.find_packages(),
)



from typing import Generator

import pytest

from dragn.dice import D4, D6, D8, D10, D12, D20
from dragn.dice.die import DieBuilder


class TestDieBuilder:
    @staticmethod
    def test_creating_one_die() -> None:
        my_new_die = DieBuilder(4)

        assert my_new_die() in range(1, 5)


class TestDie:
    @staticmethod
    def test_one_configured_die() -> None:
        assert D4() in range(1, 5)

    @staticmethod
    def test_all_configured_dice() -> None:
        configured_dice = {
            D4: range(1, 5),
            D6: range(1, 7),
            D8: range(1, 9),
            D10: range(1, 11),
            D12: range(1, 13),
            D20: range(1, 21),
        }
        for die, results_range in configured_dice.items():
            assert die() in results_range
            assert die.max_value == max(list(results_range))
            assert str(die) == f"D{die.max_value}"


@pytest.fixture
def D1() -> Generator:
    yield DieBuilder(1)


class TestDieBuilderForTumbler:
    @staticmethod
    def test_creating_tumbler_by_multiplication(D1: DieBuilder) -> None:
        tumbler = D1 * 2

        assert tumbler() == (1, 1)

    @staticmethod
    def test_creating_tumbler_by_multiplication_in_different_order(
        D1: DieBuilder
    ) -> None:
        tumbler = 2 * D1

        assert tumbler() == (1, 1)

    @staticmethod
    def test_creating_tumbler_by_addition(D1: DieBuilder) -> None:
        with pytest.raises(TypeError):
            1 + D1

    @staticmethod
    def test_creating_tumbler_by_addition_in_different_order(D1: DieBuilder) -> None:
        with pytest.raises(TypeError):
            D1 + 1

    @staticmethod
    def test_creating_tumbler_by_addition_with_another_die(D1: DieBuilder) -> None:
        tumbler = D1 + D1

        assert tumbler() == (1, 1)

    @staticmethod
    def test_creating_tumbler_by_addition_with_another_dice(D1: DieBuilder) -> None:
        tumbler = D1 + D1 + D1
        assert tumbler() == (1, 1, 1)

    @staticmethod
    def test_running_dice_with_multiplication_twice(D1: DieBuilder) -> None:
        tumbler = 3 * D6
        assert sum(tumbler()) >= 3 <= 6 * 3
        assert sum(tumbler()) >= 3 <= 6 * 3

    @staticmethod
    def test_running_dice_with_addition_twice(D1: DieBuilder) -> None:
        tumbler = D6 + D6
        assert sum(tumbler()) >= 2 <= 6 * 2
        assert sum(tumbler()) >= 2 <= 6 * 2

    @staticmethod
    def test_creating_tumbler_by_multiplying_a_tumbler(D1: DieBuilder) -> None:
        tumbler = 2 * (1 * D1)

        assert tumbler() == (1, 1)

    @staticmethod
    def test_creating_tumbler_by_multiplying_a_tumbler_other_way(
        D1: DieBuilder
    ) -> None:
        tumbler = (1 * D1) * 2

        assert tumbler() == (1, 1)

    @staticmethod
    def test_creating_tumbler_by_adding_a_tumbler(D1: DieBuilder) -> None:
        tumbler = (1 * D1) + (1 * D1)

        assert tumbler() == (1, 1)
